

//////////////////////////////////////////////////////////////////
//                                                              //
//           PLINK (c) 2005-2006 Shaun Purcell                  //
//                                                              //
// This file is distributed under the GNU General Public        //
// License, Version 2.  Please see the file COPYING for more    //
// details                                                      //
//                                                              //
//////////////////////////////////////////////////////////////////


#include <iostream>
#include <iomanip>
#include <climits>
#include <fstream>
#include <bitset>
#include <map>
#include "plink.h"
#include "options.h"
#include "perm.h"
#include "sets.h"

extern ofstream LOG;

void Plink::printLOG(string s)
{
  LOG << s;
  if (!par::silent)
    cout << s;
}

void Plink::displayGenomePV()
{

  ofstream PLO;
  string f = par::output_file_name + ".plink2";
  PLO.open(f.c_str(),ios::out);
  
  printLOG("Writing genome-wide corrected PLINK results to [ " +f +" ] \n");
  
  for (int i=0; i<original.size();i++)
    {
      for (int j=0; j<original[i].size(); j++)
	{
	  double pv=0;
	  for (int k=0; k<maxr2.size(); k++)
	    if (maxr2[k] >= original[i][j]) pv++;
	  PLO << "G " 
	      << double(pv+1)/double(par::replicates+1)
	      << "\n";
	}
    }
  PLO.close();
}

void Plink::display_listByAllele()
{

  if (!par::SNP_major)
    Ind2SNP();

  // Create an output file that lists one allele per line
  // N = number of individuals FID1 IID2 FID2 IID2 ... FIDN IIDN 
  
  // SNP ALLELE N { Individual list ( 2N entries) } 

  string f = par::output_file_name + ".recode.list";
  
  printLOG("Writing recoded list file to [ " + f + " ] \n");
  ofstream PED(f.c_str(), ios::out);
  PED.clear();

  string delimit = " ";
  if (par::recode_jes) delimit = "\t";

  
  vector<CSNP*>::iterator s = SNP.begin();
  vector<Locus*>::iterator loc = locus.begin();

  while ( s != SNP.end() )
    {
      // Genotype 11
      PED << (*loc)->chr << delimit
	  << (*loc)->name << delimit
	  << (*loc)->allele1 << (*loc)->allele1 ;
      
      vector<bool>::iterator i1 = (*s)->one.begin();
      vector<bool>::iterator i2 = (*s)->two.begin();
      vector<Individual*>::iterator gperson = sample.begin();
      
      while ( gperson != sample.end() )
	{
	  if ( (!*i1) && (!*i2) ) 
	    PED << delimit << (*gperson)->fid 
		<< delimit << (*gperson)->iid ;
	  i1++;
	  i2++;
	  gperson++;
	}
      PED << "\n";
      
      // Genotype 12
      PED << (*loc)->chr << delimit
	  << (*loc)->name << delimit
	  << (*loc)->allele1 << (*loc)->allele2 ;
      i1 = (*s)->one.begin();
      i2 = (*s)->two.begin();
      gperson = sample.begin();
      while ( gperson != sample.end() )
	{
	  if ( (!*i1) && *i2 ) 
	    PED << delimit << (*gperson)->fid 
		<< delimit << (*gperson)->iid ;
	  i1++;
	  i2++;
	  gperson++;
	}
      PED << "\n";

      // Genotype 22
      PED << (*loc)->chr << delimit
	  << (*loc)->name << delimit
	  << (*loc)->allele2 << (*loc)->allele2 ;
      i1 = (*s)->one.begin();
      i2 = (*s)->two.begin();
      gperson = sample.begin();
      while ( gperson != sample.end() )
	{
	  if ( *i1 && *i2 ) 
	    PED << delimit << (*gperson)->fid 
		<< delimit << (*gperson)->iid ;
	  i1++;
	  i2++;
	  gperson++;
	}
      PED << "\n";

      // Genotype 00
      PED << (*loc)->chr << delimit
	  << (*loc)->name << delimit
	  << "00";
      i1 = (*s)->one.begin();
      i2 = (*s)->two.begin();
      gperson = sample.begin();
      while ( gperson != sample.end() )
	{
	  if ( *i1 && !*i2 ) 
	    PED << delimit << (*gperson)->fid 
		<< delimit << (*gperson)->iid ;
	  i1++;
	  i2++;
	  gperson++;
	}
      PED << "\n";
      
      
      // Next SNP
      s++;
      loc++;
    }
  
  PED.close();

}


void Plink::extractExcludeSet(bool exclude)
{
  
  // Make map of locus name with 'l' number
  map<string,int> mlocus;
  vector<bool> del(nl_all);
  for (int l=0;l<nl_all;l++)
    {
      mlocus.insert(make_pair(locus[l]->name,l));
      if (exclude) del[l] = false;  // start off all included
      else del[l] = true;           // start off all excluded
    }
    
  map<string,int>::iterator ilocus;
  
  
  //////////////////////////////////
  // Either extract a certain "GENE"

  if (par::dump_gene)
    {
      
      readSet();
      Set S(snpset);
      
      bool found_gene = false;
      int c=0;

      for (int s=0; s<setname.size(); s++)
	{

	  if ( setname[s] == par::dump_genename )
	    {
	      // Retain just these SNPs
	      for (int l=0; l<snpset[s].size(); l++)
		del[snpset[s][l]] = false;
	      found_gene = true;
	      c = snpset[s].size();
	      break;
	    }
	  if (found_gene) break;
	}
      
      if (found_gene)
	printLOG(int2str(c)+" SNPs selected from [ " + par::dump_genename + " ]\n");
      else
	printLOG("WARNING: ** Could not find [ " + par::dump_genename + " ] in [ " + par::setfile + " ]\n");

    }

  /////////////////////////////////////
  // OR get a list of SNPs from a file

  else
    {

      checkFileExists(par::dump_set_file);  
      ifstream INFILE(par::dump_set_file.c_str(),ios::in);
      INFILE.clear();

      printLOG("Reading list of SNPs ");
      if (exclude) printLOG("to exclude [ ");
      else printLOG("to extract [ ");
      printLOG(par::dump_set_file + " ] ... ");
      
      int c=0;
      while (!INFILE.eof())
	{
	  string m;
	  INFILE >> m;
	  if (m=="") continue;
	  
	  ilocus = mlocus.find(m);
	  if (ilocus != mlocus.end())
	    {
	      if (exclude) del[ilocus->second] = true; 
	      else del[ilocus->second] = false; 
	      c++;
	    }
	  else
	    {
	      printLOG("\n"+m+" in [ "+
		       par::dump_set_file+" ] not in [ "+
		       par::mapfile +" ]\n");
	    }
	}
      INFILE.close();
      printLOG(int2str(c)+" read\n");
    }


  ////////////////////////////////////////
  // Remove selected loci from locus list, 
  
  deleteSNPs(del);

  // Re-Read sets, as the SNP numbers will now have changed
  if (par::dump_gene)
    readSet();

}

void Plink::removeIndividuals(bool keep)
{
  

  //////////////////////////////////
  // Make map of individuals FID/IID

  map<string,int> mperson;
  map<string,int>::iterator iperson;
  vector<bool> del(n);
  for (int i=0;i<n;i++)
    {
      mperson.insert(make_pair(sample[i]->fid+"_"+sample[i]->iid,i));
      if (!keep) del[i] = false;    // start off all included
      else del[i] = true;           // start off all excluded
    }
  

  ///////////////////////////
  // Read list of individuals

  checkFileExists(par::remove_indiv_list);  
  ifstream INFILE(par::remove_indiv_list.c_str(),ios::in);
  INFILE.clear();
   
  printLOG("Reading individuals ");
  if (keep) printLOG(" to keep [ ");
  else printLOG( " to remove [ ");
  printLOG( par::remove_indiv_list + " ] ... ");

  int c=0;
  while (!INFILE.eof())
    {
      string fid,iid;
      INFILE >> fid >> iid;
      if (fid=="" || iid=="") continue;
      
      iperson = mperson.find(fid+"_"+iid);
      if (iperson != mperson.end())
	{
	  if (!keep) del[iperson->second] = true; 
	  else del[iperson->second] = false; 
	  c++;
	}
      else
	{
	  printLOG( "\n" + fid + " " + iid + " in [ " + par::remove_indiv_list + 
		    " ] not in [ " + par::pedfile + " ]\n");
	}
    }
  INFILE.close();
  printLOG(int2str(c)+" read\n");
  


  ////////////////////////////////////
  // Remove individuals as appropriate

  int n_removed = deleteIndividuals(del);

  printLOG(int2str(n_removed)+" individuals removed with ");
  if (keep) printLOG("--keep option\n");
  else printLOG("--remove option\n");
  
}


void Plink::flipStrand()
{

  ////////////////////////////
  // Look-up table by SNP name

  map<string,Locus*> mlocus;
  map<string,Locus*>::iterator ilocus;

  vector<Locus*>::iterator loc = locus.begin();
  while ( loc != locus.end() )
    {
      mlocus.insert(make_pair( (*loc)->name, *loc) );
      loc++;
    }  

  ///////////////////////
  // Read in SNPs to flip

  checkFileExists(par::flip_file);  
  printLOG("Reading SNPs to flip strand from [ " + par::flip_file + " ] \n");
  ifstream INFILE(par::flip_file.c_str(),ios::in);
  INFILE.clear();
  int counter = 0;

  while (!INFILE.eof())
    {
      string m;
      INFILE >> m;
      if (m=="") continue;
      
      ilocus = mlocus.find(m);
      if (ilocus != mlocus.end())
	{
	  counter++;
	  // Flip strand
	  
	  Locus * loc = ilocus->second;
	  
	  if ( loc->allele1 == "A" ) loc->allele1 = "T";
	  else if ( loc->allele1 == "C" ) loc->allele1 = "G";
	  else if ( loc->allele1 == "G" ) loc->allele1 = "C";
	  else if ( loc->allele1 == "T" ) loc->allele1 = "A";
	  else if ( loc->allele1 == "1" ) loc->allele1 = "4";
	  else if ( loc->allele1 == "2" ) loc->allele1 = "3";
	  else if ( loc->allele1 == "3" ) loc->allele1 = "2";
	  else if ( loc->allele1 == "4" ) loc->allele1 = "1";
	  
	  if ( loc->allele2 == "A" ) loc->allele2 = "T";
	  else if ( loc->allele2 == "C" ) loc->allele2 = "G";
	  else if ( loc->allele2 == "G" ) loc->allele2 = "C";
	  else if ( loc->allele2 == "T" ) loc->allele2 = "A";
	  else if ( loc->allele2 == "1" ) loc->allele2 = "4";
	  else if ( loc->allele2 == "2" ) loc->allele2 = "3";
	  else if ( loc->allele2 == "3" ) loc->allele2 = "2";
	  else if ( loc->allele2 == "4" ) loc->allele2 = "1";
	  
	}

    }

  INFILE.close();
  
  printLOG("Flipped strand of " + int2str(counter) + " SNPs\n");
  
}


void Plink::display_recoded_PEDFILE()
{
  
  string f = par::output_file_name + ".recode.ped";
  
  printLOG("Writing recoded ped file to [ " + f + " ] \n");
  ofstream PED(f.c_str(), ios::out);
  PED.clear();

  string delimit = " ";
  if (par::recode_jes) delimit = "\t";
  
  for (int i=0;i<n;i++)
    {
      Individual * person = sample[i];
      PED << person->fid << delimit
	  << person->iid << delimit
	  << person->pat << delimit
	  << person->mat << delimit
	  << person->sexcode << delimit;

      if (par::bt)
	PED << (int)person->phenotype;
      else
	PED << person->phenotype;
      
      
      for (int l=0;l<nl_all;l++)
	{

	  if (par::SNP_major)
	    {
	      if      ( (!SNP[l]->one[i]) && (!SNP[l]->two[i]) )  
		PED << delimit << locus[l]->allele1 << " " << locus[l]->allele1;
	      else if ( (!SNP[l]->one[i]) && SNP[l]->two[i]) 
		PED << delimit << locus[l]->allele1 << " " << locus[l]->allele2;
	      else if (  SNP[l]->one[i]   && SNP[l]->two[i]) 
		PED << delimit << locus[l]->allele2 << " " << locus[l]->allele2;
	      else PED << delimit << par::missing_genotype << " " << par::missing_genotype;
	    }
	  else
	    {
	      if      ( (!person->one[l]) && (!person->two[l]) )  
		PED << delimit << locus[l]->allele1 << " " << locus[l]->allele1;
	      else if ( (!person->one[l]) && person->two[l]) 
		PED << delimit << locus[l]->allele1 << " " << locus[l]->allele2;
	      else if (  person->one[l]   && person->two[l]) 
		PED << delimit << locus[l]->allele2 << " " << locus[l]->allele2;
	      else PED << delimit << par::missing_genotype << " " << par::missing_genotype;
	    }

	}
      PED << "\n";
    }
  
  PED.close();

  // And we also need a new map file
  
  f = par::output_file_name + ".recode.map";
  printLOG( "Writing new map file to [ " + f + " ] \n");
  ofstream MAP(f.c_str(), ios::out);
  MAP.clear();
  
  for (int l=0;l<nl_all;l++)
    {
      MAP << locus[l]->chr << "\t"
	  << locus[l]->name << "\t"
	  << locus[l]->pos << "\t"
	  << locus[l]->bp << "\n";
    }
  
  MAP.close();
  
}

void Plink::display_recoded_PEDFILE_12()
{

  string f = par::output_file_name + ".recode.ped";
  printLOG( "Writing recoded ped file to [ " + f + " ] \n");
  ofstream PED(f.c_str(), ios::out);
  PED.clear();

  string delimit = " ";
  if (par::recode_jes) delimit = "\t";
  
  for (int i=0;i<n;i++)
    {
      Individual * person = sample[i];
      PED << person->fid << delimit
	  << person->iid << delimit
	  << person->pat << delimit
	  << person->mat << delimit
	  << person->sexcode << delimit;
      
      if (par::bt)
	PED << (int)person->phenotype;
      else
	PED << person->phenotype;

      for (int l=0;l<nl_all;l++)
	{
	  
	  if (par::SNP_major)
	    {
	      if      ( (!SNP[l]->one[i]) && (!SNP[l]->two[i]) )  
		PED << delimit << "1 1";
	      else if ( (!SNP[l]->one[i]) && SNP[l]->two[i]) 
		PED << delimit << "1 2";
	      else if (  SNP[l]->one[i]   && SNP[l]->two[i]) 
		PED << delimit << "2 2";
	      else PED << delimit << "0 0";
	    }
	  else
	    {
	      if      ( (!person->one[l]) && (!person->two[l]) )  
		PED << delimit << "1 1";
	      else if ( (!person->one[l]) && person->two[l]) 
		PED << delimit << "1 2";
	      else if (  person->one[l]   && person->two[l]) 
		PED << delimit << "2 2";
	      else PED << delimit << "0 0";
	    }
	  
	}
      PED << "\n";
    }

  PED.close();

  // And we also need a new map file

  f = par::output_file_name + ".recode.map";
  printLOG( "Writing new map file to [ " + f + " ] \n");  
  ofstream MAP(f.c_str(), ios::out);
  MAP.clear();
  
  for (int l=0;l<nl_all;l++)
    {
      MAP << locus[l]->chr << "\t"
 	  << locus[l]->name << "\t"
 	  << locus[l]->pos << "\t"
	  << locus[l]->bp << "\n";
    }
  MAP.close();
  
}

void Plink::display_recoded_PEDFILE_AD()
{

  string f = par::output_file_name + ".recode.raw";
  printLOG( "Writing recoded file to [ " + f + " ] \n");
  ofstream PED(f.c_str(), ios::out);
  PED.clear();
  
  // Header row
  PED << "FID IID PAT MAT SEX PHENOTYPE ";
  for (int l=0;l<nl_all;l++)
    PED << locus[l]->name+"_A" << " " 
	<< locus[l]->name+"_D" << " "
	<< " ";
  PED << "\n";

  for (int i=0;i<n;i++)
    {
      Individual * person = sample[i];
      PED << person->fid << " "
	  << person->iid << " "
	  << person->pat << " "
	  << person->mat << " "
	  << person->sexcode << " ";

      if (par::bt)
	PED << (int)person->phenotype << " ";
      else
	PED << person->phenotype << " ";

      for (int l=0;l<nl_all;l++)
	{
	  
	  if (par::SNP_major)
	    {
	      
	      if (par::recode_AD_fixed && locus[l]->allele1=="1")
		{
		  if      ( (!SNP[l]->one[i]) && (!SNP[l]->two[i]) )  
		    PED << "2 0 ";
		  else if ( (!SNP[l]->one[i]) && SNP[l]->two[i]) 
		    PED << "1 1 ";
		  else if (  SNP[l]->one[i]   && SNP[l]->two[i]) 
		    PED << "0 0 ";
		  else PED << "NA NA ";
		}
	      else
		{
		  if      ( (!SNP[l]->one[i]) && (!SNP[l]->two[i]) )  
		    PED << "0 0 ";
		  else if ( (!SNP[l]->one[i]) && SNP[l]->two[i]) 
		    PED << "1 1 ";
		  else if (  SNP[l]->one[i]   && SNP[l]->two[i]) 
		    PED << "2 0 ";
		  else PED << "NA NA ";
		}
	    }
	  else
	    {
	      
	      if (par::recode_AD_fixed && locus[l]->allele1=="1")
		{
		  if      ( (!person->one[l]) && (!person->two[l]) )  
		    PED << "2 0 ";
		  else if ( (!person->one[l]) && person->two[l]) 
		    PED << "1 1 ";
		  else if (  person->one[l]   && person->two[l]) 
		    PED << "0 0 ";
		  else PED << "NA NA ";
		}
	      else
		{
		  if      ( (!person->one[l]) && (!person->two[l]) )  
		    PED << "0 0 ";
		  else if ( (!person->one[l]) && person->two[l]) 
		    PED << "1 1 ";
		  else if (  person->one[l]   && person->two[l]) 
		    PED << "2 0 ";
		  else PED << "NA NA ";
		}

	      
	    }
	  

	}
      PED << "\n";
    }

  PED.close();

}


void Plink::write_BITFILE()
{

  printLOG( "Writing pedigree information to [ " 
	    + par::output_file_name + ".fam ] \n");
  ofstream BIT((par::output_file_name+".fam").c_str(), ios::out);
  // For each individual
  for (int i=0;i<n;i++)
    {
      Individual * person = sample[i];
      BIT << person->fid << " "
	  << person->iid << " "
	  << person->pat << " "
	  << person->mat << " "
	  << person->sexcode << " ";

      if (par::bt)
	BIT << (int)person->phenotype << "\n";
      else
	BIT << person->phenotype << "\n";
    }
  BIT.clear();
  BIT.close();

  printLOG( "Writing map (extended format) information to [ " 
	    + par::output_file_name + ".bim ] \n");
  BIT.open((par::output_file_name+".bim").c_str(), ios::out);
  
  for (int l=0;l<nl_all;l++)
    {

      if (locus[l]->allele1=="")
 	{
 	  if (locus[l]->allele2!="0") locus[l]->allele1="0";
 	  else locus[l]->allele1="X";
 	}
      
      if (locus[l]->allele2=="")
	locus[l]->allele2="0";
      
      BIT << locus[l]->chr << "\t"
	  << locus[l]->name << "\t"
	  << locus[l]->pos << "\t"
	  << locus[l]->bp << "\t" 
	  << locus[l]->allele1 << "\t"
	  << locus[l]->allele2 << "\n";
    }
  
  BIT.clear();
  BIT.close();


  //////////////////////////////////////
  // Save genotype data in BITFILE format

  printLOG("Writing genotype bitfile to [ " 
	   + par::output_file_name + ".bed ] \n");
    
  BIT.open((par::output_file_name+".bed").c_str(), ios::out | ios::binary);
  
  if (par::out_SNP_major)
    printLOG("Using (default) SNP-major mode\n");
  else
    printLOG("Using individual-major mode\n");

  bitset<8> b;
  char ch[1];


  // Magic numbers for .bed file: 00110110 11011000 = v1.00 bed file

  b.reset();
  b.set(2);  b.set(3);  b.set(5);  b.set(6);
  ch[0] = (char)b.to_ulong();
  BIT.write(ch,1);

  b.reset();
  b.set(0);  b.set(1);  b.set(3);  b.set(4);
  ch[0] = (char)b.to_ulong();
  BIT.write(ch,1);


  // BIT represents status of SNP-major (true) or Ind-major (false)

  b.reset();
  if (par::out_SNP_major)
    b.set(0);
  ch[0] = (char)b.to_ulong();
  BIT.write(ch,1);
  
  
  //////////////////////////////////////////
  // Now consider genotypes: SNP-major mode
  
  if (par::out_SNP_major)
    {  
      
      // Create the SNP-major dataset, if need be
      if (!par::SNP_major) Ind2SNP();

      vector<CSNP*>::iterator s = SNP.begin();
         
      // Outer loop over SNPs
      while ( s != SNP.end() )
	{
	  
	  vector<bool>::iterator i1 = (*s)->one.begin();
	  vector<bool>::iterator i2 = (*s)->two.begin();
	  //	  vector<Individual*>::iterator person = sample.begin();

	  // Inner loop over individuals
	  while ( i1 != (*s)->one.end() )
	    {
	      bitset<8> b;
	      b.reset();
	      int c=0;      
	      
	      while (c<8 && i1 != (*s)->one.end() )
		{
		  if ( *(i1) ) b.set(c);
		  i1++;
		  c++;

		  if ( *(i2) ) b.set(c);		  
		  i2++;
		  c++;

		  //person++;
		}
	      
	      char ch[1];
	      ch[0] = (char)b.to_ulong();
	      BIT.write(ch,1);
	      
	    }

	  // next SNP
	  s++;
	}
    }
  

  ////////////////////////////////////////////////////////////
  // Alternatively, consider genotypes: Individual-major mode

  else
    {

      // Create the individual-major dataset, if need be
      if (par::SNP_major) SNP2Ind();
      
      vector<Individual*>::iterator person = sample.begin();
            
      // Outer loop over individuals
      while ( person != sample.end() )
	{
	  
	  vector<bool>::iterator i1 = (*person)->one.begin();
	  vector<bool>::iterator i2 = (*person)->two.begin();
	  
	  // Inner loop over SNPs
	  while ( i1 != (*person)->one.end() )
	    {
	      
	      bitset<8> b;
	      b.reset();
	      int c=0;      
	      
	      while (c<8 && i1 != (*person)->one.end() )
		{
		  if ( *(i1) ) b.set(c);
		  i1++;
		  c++;
		  if ( *(i2) ) b.set(c);		  
		  i2++;
		  c++;
		}
	      
	      char ch[1];
	      ch[0] = (char)b.to_ulong();
	      BIT.write(ch,1);
	      
	    }
	  
	  // next person
	  person++;
	}
      

    }


  BIT.close();
  
}



void Plink::summaryIBSsegments(Perm & perm)
{

  vector<int> coverage_conc_aff(nl_all,0);
  vector<int> coverage_conc_unaff(nl_all,0);
  vector<int> coverage_disc(nl_all,0);
    
  vector<Segment>::iterator s = segment.begin();
  
  while ( s != segment.end() )
    {    
      
      if (s->p1->aff == s->p2->aff)
	{
	  if ( s->p1->aff) 
	    for (int l = s->start ; l <= s->finish; l++) coverage_conc_aff[l]++;
	  else
	    for (int l = s->start ; l <= s->finish; l++) coverage_conc_unaff[l]++;
	}
      else
	for (int l = s->start ; l <= s->finish; l++) coverage_disc[l]++;
              
      s++;
    }
    
  
  string f = par::output_file_name;
  if ( par::ibs_2only )
    f += ".ibs2.summary";
  else if ( par::segment_output ) 
    f += ".segment.summary";
  else
    f += ".ibs.summary";
  
  ofstream SIBS;
  SIBS.open( f.c_str() , ios::out );
  

  //////////////////////////////
  // And display
  
  for (int l=0; l<nl_all; l++)
    {
      
      SIBS << setw(4) << locus[l]->chr << " " 
	   << setw(par::pp_maxsnp) << locus[l]->name << " " 
	   << setw(8) << coverage_conc_unaff[l] << " " 
	   << setw(8) << coverage_disc[l] << " " 
	   << setw(8) << coverage_conc_aff[l] << "\n"; 
      
    }

  SIBS.close();


  //////////////////////////////////
  // Permutation test, if C/C given?

  if (!par::permute) return;
  
  segmentPermutationTest(perm,false,f,coverage_conc_aff,coverage_disc,coverage_conc_unaff);

  return;

}


void Plink::summaryIBDsegments(Perm & perm)
{

  // Number of map positions
  // (extra position for global IBD)
  int npos = m1.size() - 1;
  
  vector<int> coverage_conc_aff(nl,0);
  vector<int> coverage_conc_unaff(nl,0);
  vector<int> coverage_disc(nl,0);
  
  // pihat[pair][position]
  // check if, e.g. pihat>0.2 (i.e. atleast 80% chance of IBD1 sharing)
  // level set in options.cpp

      
  for (int l=0;l<npos;l++)
    {
      
      double p1, p2;
    
      if (m1[l]==-1) 
	{
	  p1 = locus[par::run_start]->pos - par::fringe;
	}
      else 
	{
	  p1 = locus[m1[l]]->pos;
	}
      
      if (m2[l]==-1) 
	{
	  p2 = locus[par::run_end]->pos + par::fringe;
	}
      else 
	{
	  p2 = locus[m2[l]]->pos;
	}
      
      if (m1[l]==-1 && m2[l]==-1)
	{
	  p1 = p2 = 0;
	}
      
      double d1 = p1 + pos[l] * (p2-p1);

      
      vector<Segment>::iterator s = segment.begin();
      
      while ( s != segment.end() )
	{    
	  
	  if ( locus[s->start]->chr != par::run_chr )
	    {
	      s++;
	      continue;
	    }
	  
	  
	  // In segment?
	  if (locus[s->start]->pos < d1 && locus[s->finish]->pos > d1  )
	    {
	      
	      if (s->p1->aff == s->p2->aff)
		{
		  if ( s->p1->aff) 
		    coverage_conc_aff[l]++;
		  else
		    coverage_conc_unaff[l]++;
		}
	      else
		coverage_disc[l]++;
	      
	    }
	  
	  s++;
	}
      
    }
  
  
  string f = par::output_file_name + ".segment.summary";

  ofstream SSEG;
  if (par::segment_output_started)
     SSEG.open( f.c_str() , ios::app );
  else
     SSEG.open( f.c_str() , ios::out );
  par::segment_output_started = true;

  //////////////////////////////
  // And display
  
  for (int l=0; l<npos; l++)
    {
      
      double p1, p2;
      string n1, n2;
      double fr=0;
      
      ////////////////////////////////
      // Single and multipoint output
      
      if (m1[l]==-1) 
	{
	  p1 = locus[par::run_start]->pos - par::fringe;
	  n1 = "fringe";
	}
      else 
	{
	  p1 = locus[m1[l]]->pos;
	  n1 = locus[m1[l]]->name;
	  fr = locus[m1[l]]->freq;
	}
      
      if (m2[l]==-1)
	{
	  p2 = locus[par::run_end]->pos + par::fringe;
	  n2 = "fringe";
	}
      else 
	{
	  p2 = locus[m2[l]]->pos;
	  n2 = locus[m2[l]]->name;
	}
      
      if (m1[l]==-1 && m2[l]==-1)
	{
	  p1 = p2 = 0;
	  n1 = "Genomewide";
	  n2 = "IBD";
	}

      double d1 = p1 + pos[l] * (p2-p1);

      SSEG << par::run_chr << " " 
	   << n1 << " " 
	   << n2 << " " 
	   << d1 << "  "
	   << fr << " " 
	   << coverage_conc_unaff[l] << " " 
	   << coverage_disc[l] << " " 
	   << coverage_conc_aff[l] << "\n"; 

    }

  SSEG.close();

  
  ////////////////////////
  // C/C permutation test?

  if (!par::permute) return;

  segmentPermutationTest(perm,true,f,coverage_conc_aff,coverage_disc,coverage_conc_unaff);

}



void Plink::summaryIBD()
{

  int npos = pihat[0].size();
  int npair = pihat.size();
  
  vector<int> coverage(npos,0);
  
  vector<int> coverage_conc_aff(npos,0);
  vector<int> coverage_conc_unaff(npos,0);
  vector<int> coverage_disc(npos,0);
  
  // pihat[pair][position]
  // check if, e.g. pihat>0.2 (i.e. atleast 80% chance of IBD1 sharing)
  // level set in options.cpp

  for (int i=0;i<pihat.size();i++)
    for (int j=0;j<npos;j++)
      {
	if (pihat[i][j] > par::IBD_threshold )
	  {
	    coverage[j]++;
	    
	    if (sample[pair1[i]]->phenotype != sample[pair2[i]]->phenotype) 
	      coverage_disc[j]++;

	    if (sample[pair1[i]]->phenotype==1 && sample[pair2[i]]->phenotype==1) 
	      coverage_conc_unaff[j]++;

	    if (sample[pair1[i]]->phenotype==2 && sample[pair2[i]]->phenotype==2) 
	      coverage_conc_aff[j]++;
	  }
      }


  //////////////////////////////
  // And display
  
  for (int l=0; l<npos; l++)
    {
      
      double p1, p2;
      string n1, n2;
      double fr = 0;

      ////////////////////////////////
      // Single and multipoint output
      
      if (m1[l]==-1) 
	{
	  p1 = locus[par::run_start]->pos - par::fringe;
	  n1 = "fringe";
	}
      else 
	{
	  p1 = locus[m1[l]]->pos;
	  n1 = locus[m1[l]]->name;
	  fr = locus[m1[l]]->freq;
	}
      
      if (m2[l]==-1) 
	{
	  p2 = locus[par::run_end]->pos + par::fringe;
	  n2 = "fringe";
	}
      else 
	{
	  p2 = locus[m2[l]]->pos;
	  n2 = locus[m2[l]]->name;
	}
      
      if (m1[l]==-1 && m2[l]==-1)
	{
	  p1 = p2 = 0;
	  n1 = "Genomewide";
	  n2 = "IBD";
	}
      
      double d1 = p1 + pos[l] * (p2-p1);
      
      cout << "C " 
	   << par::run_chr << " " 
	   << n1 << " " 
	   << n2 << " " 
	   << d1 << "  "
	   << fr << " " 
	   << coverage[l] << " " 
	   << coverage_conc_unaff[l] << " " 
	   << coverage_disc[l] << " " 
	   << coverage_conc_aff[l] << " " 
	   << coverage[l]/(double)npair << "\n"; 
    }
}


void Plink::displayGMULTI(Individual * p1, Individual * p2, int l, ofstream & GMULTI)
{

  Locus * loc = locus[l];

  bool a1 = p1->one[l];
  bool a2 = p1->two[l];

  bool b1 = p2->one[l];
  bool b2 = p2->two[l];

  bool miss = false;
  
  GMULTI << setw(par::pp_maxfid) << p1->fid << " "
	 << setw(par::pp_maxiid) << p1->iid << " "
	 << setw(par::pp_maxfid) << p2->fid << " "
	 << setw(par::pp_maxiid) << p2->iid << " "
	 << setw(par::pp_maxsnp) << loc->name << " ";

  // Minor allele and frequency 

  GMULTI << setw(2) << loc->allele1 << " "
	 << setw(8) << loc->freq << " ";

  // Genotypes

  if ((!a1) && (!a2)) 
    GMULTI << setw(4) 
	   << loc->allele1+"/"+loc->allele1 
	   << " ";
  else if ((!a1) && a2) 
    GMULTI << setw(4) 
	   << loc->allele1+"/"+loc->allele2
	   << " ";
  else if (a1 && a2) 
    GMULTI << setw(4) 
	   << loc->allele2+"/"+loc->allele2
	   << " ";
  else 
    {
      GMULTI << setw(4) << "0/0" << " ";
      miss = true; 
    }
  
  if ((!b1) && (!b2)) 
    GMULTI << setw(4) 
	   << loc->allele1+"/"+loc->allele1;
  else if ((!b1) && b2) 
    GMULTI << setw(4) 
	   << loc->allele1+"/"+loc->allele2;
  else if (b1 && b2) 
    GMULTI << setw(4) 
	   << loc->allele2+"/"+loc->allele2;
  else 
    {
      GMULTI << setw(4) << "0/0";
      miss = true;
    }

  // IBS count

  if (miss)
    GMULTI << setw(3) << "NA" << " ";
  else
    {
      int ibs=0;
      if (a1==b1) ibs++;
      if (a2==b2) ibs++;
      GMULTI << setw(3) << ibs << " ";
    }

  GMULTI << "\n";
  
}
